// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_ClientSocket_hpp
#define anna_comm_ClientSocket_hpp

#include <anna/comm/Socket.hpp>
#include <anna/comm/Buffer.hpp>
#include <anna/comm/Transport.hpp>
#include <anna/core/util/Millisecond.hpp>

namespace anna {

namespace comm {

class Communicator;
class DatagramSocket;
class Transport;
class Message;
class Server;
class CongestionController;
class Receiver;

namespace handler {
class MetaClientSocket;
class DatagramSocket;
}

/**
   Implements client sockets (aka @ref Socket) established by our application with a remote endpoint
   where a ServerSocket is accepting connection requests.
*/
class ClientSocket : public Socket {
public:
  /**
     Numero de milisegundos por defecto que espera antes de dar por fallida una conexion con
     otro proceso servidor.
  */
  static const Millisecond DefaultMaxConnectionDelay;

  /**
     Numero de milisegundos por defecto que queda bloqueado un proceso a la espera de poder
     escribir en un socket cuyo buffer de salida esta lleno.
  */
  static const Millisecond DefaultMaxWriteDelay;

  /**
     Crea un socket cliente liberado.
     \param transportFactory factoria de protocolos de transporte a usar por este sockets.
     \param domain Dominio del socket.
     \param type Tipo de socket.
     \warning La factoria de protocolos debe estar disponible mientras el Socket esta activo.
  */
  ClientSocket(TransportFactory* transportFactory = NULL, Domain::_v domain = Socket::Domain::Inet, Type::_v type = Socket::Type::Stream) :
    Socket(domain, type, transportFactory),
    a_data(true),
    a_reader(true),
    a_ignoreIncomingMessages(false) {
    initialize();
  }

  /**
     Crea un socket cliente que sera conectado remotamente a la direccion y puerto indicado.

     \param remoteAddress direccion de red remota a la que conectar.
     \param transportFactory factoria de protocolos de transporte a usar por este sockets.
     \param type Tipo de socket.
     \warning La factoria de protocolos debe estar disponible mientras el Socket esta activo.
  */
  ClientSocket(const INetAddress& remoteAddress, TransportFactory* transportFactory = NULL, const Type::_v type = Socket::Type::Stream) :
    Socket(Socket::Domain::Inet, type, transportFactory),
    a_remoteAccessPoint(remoteAddress),
    a_data(true),
    a_reader(true),
    a_ignoreIncomingMessages(false) {
    initialize();
  }

  /**
     Crea un socket cliente que sera conectado al archivo indicado.

     \param path Ruta del archivo que vamos a usar para transferir datos ataves de este socket.
     \param type Tipo de socket.
  */
  ClientSocket(const std::string& path, const Type::_v type = Socket::Type::Stream) :
    Socket(path, type),
    a_remoteAccessPoint(path),
    a_data(true),
    a_reader(true),
    a_ignoreIncomingMessages(false) {
    initialize();
  }

  /**
     Crea un socket cliente conectado al servidor indicado por la direccion y puerto remoto. Ademas el socket cliente
     seria conectado localmente (ver @ref Socket::bind) a la direccion y puerto locales indicados.

     \param transportFactory factoria de protocolos de transporte a usar por este sockets.
     \param remoteAddress direccion del servidor.
     \param localAddress Puede ser usado para limitar la direccion por la que atendiende peticiones un servidor de socket
     instalado en una maquina con mas de una direccion.
     compartido por mas de un proceso activo.
     \param type Tipo de socket.
     \warning La factoria de protocolos debe estar disponible mientras el Socket esta activo.
  */
  ClientSocket(const INetAddress& remoteAddress,  const INetAddress& localAddress, TransportFactory* transportFactory = NULL, const Type::_v type = Socket::Type::Stream) :
    Socket(localAddress, type, transportFactory),
    a_remoteAccessPoint(remoteAddress),
    a_data(true),
    a_reader(true),
    a_ignoreIncomingMessages(false) {
    initialize();
  }

  /**
     Destructor.
  */
  virtual ~ClientSocket() { close(); }

  /**
     Devuelve la direccion remota del socket.
     \return La direccion remota del socket.
  */
  const AccessPoint& getRemoteAccessPoint() const throw() { return a_remoteAccessPoint; }

  /**
     Devuelve el estado de conexion de este socket.
     \return \em true si el socket mantiene la conexion realizada mediante el metodo #connect o
     \em false en otro caso.
  */
  bool isConnected() const throw() { return (a_status & Status::Connected) != 0; }

  /**
     Devuelve el estado del contenido del socket.
     \return \em true si el socket mantiene el sincronismo o \em false en otro caso.
  */
  bool isSynchronized() const throw() { return (a_status & Status::Corrupt) == 0; }

  /**
     Devuelve el estado del contenido del socket.
     \return \em true si el socket NO mantiene el sincronismo o \em false en otro caso.
  */
  bool isCorrupt() const throw() { return (a_status & Status::Corrupt); }

  /**
   * Devuelve el estado de cierre del socket.
   * \return \em true si el socket est a la espera de ser cerrado o \em false en otro caso.
   */
  bool isClosedPending() const throw() { return (a_status & Status::ClosePending) != 0; }

  /**
     Devuelve la instancia de anna::comm::Server asociado a una conexion remota iniciada por
     nuestra aplicacion.
     \return la instancia de anna::comm::Server asociado a una conexion remota iniciada por
     nuestra aplicacion.
     \warning Puede ser NULL en caso de que este ClientSocket corresponda a una conexion local
     establecida por un proceso remoto contra un ServerSocket de nuestra aplicacion.
  */
  Server* getServer() throw(RuntimeException);

  /**
     Devuelve el numero maximo de milisegundos esperados para obtener conexion al invocar
     al metodo #connect.
     \return el numero maximo de milisegundos esperados para obtener conexion al invocar
     al metodo #connect.
  */
  const Millisecond &getMaxConnectionDelay() const throw() { return a_msMaxConnectionDelay; }

  /**
     Devuelve el numero maximo de milisegundos que queda bloqueado el proceso/thread a la espera
     de escribir en un socket cuyo buffer de salida esta lleno.
     \return Devuelve el numero maximo de milisegundos que queda bloqueado el proceso/thread a la espera
     de escribir en un socket cuyo buffer de salida esta lleno.
  */
  const Millisecond &getMaxWriteDelay() const throw() { return a_msMaxWriteDelay; }

  /**
     Obtiene toda la informacion referente a este socket a partir de la conexion realizada atraves del \em fd
     recibido como parametro.
     \param fd File descriptor correspondiente a una conexion local.
     \warning Exclusivamente uso interno.
  */
  virtual void setfd(const int fd) throw(RuntimeException);

  /**
     Establece el numero maximo de milisegundos esperados para obtener la conexion al
     invocar al metodo #connect.
     \param msMaxConnectionDelay Numero de milisegundos esperados para obtener conexion.

     \see anna::comm::Server::setMaxConnectionDelay.
  */
  void setMaxConnectionDelay(const Millisecond &msMaxConnectionDelay)
  throw() {
    a_msMaxConnectionDelay = msMaxConnectionDelay;
  }

  /**
     Establece el numero maximo de milisegundos que queda bloqueado el proceso/thread a la espera
     de escribir en un socket cuyo buffer de salida esta lleno.

     \param msMaxWriteDelay Numero de milisegundos esperados en caso de que el buffer del socket se llene.
  */
  void setMaxWriteDelay(const Millisecond &msMaxWriteDelay) throw() { a_msMaxWriteDelay = msMaxWriteDelay; }

  /**
   * Devuelve \em true si el indicador que ignora los mensajes entrantes est activo, o \em false en otro caso.
   * \return \em true si el indicador que ignora los mensajes entrantes est activo, o \em false en otro caso.
   */
  bool getIgnoreIncomingMessages() const throw() { return a_ignoreIncomingMessages; }

  /**
   * Establece el indicador que provoca ignorar los mensajes entrantes.
   * \param ignoreIncomingMessages \em true si el indicador que ignora los mensajes entrantes est activo, o \em false en otro caso.
   */
  void setIgnoreIncomingMessages(const bool ignoreIncomingMessages) throw() { a_ignoreIncomingMessages = ignoreIncomingMessages; }

  /**
     Intenta la conexion remota con la direccion indicada en el constructor.
  */
  virtual void connect() throw(RuntimeException);

  /**
     Envia el mensaje recibido como parametro. Al bloque de datos correspondiente a la
     codificacion del mensaje se incorpora la informacion necesaria para el protocolo
     de la capa de transporte indicado en el constuctor.

     \param message Mensaje que vamos codificar para enviar a la capa de transporte.
  */
  void send(Message& message) throw(RuntimeException);

  /**
     Envia el mensaje recibido como parametro. Al bloque de datos correspondiente a la
     codificacion del mensaje se incorpora la informacion necesaria para el protocolo
     de la capa de transporte indicado en el constuctor.

     \param message Mensaje que vamos codificar para enviar a la capa de transporte.
  */
  void send(Message* message) throw(RuntimeException);

  /**
     Espera la llegada de mensajes por este ClientSocket durante un numero de milisegundos
     recibido como parametro o hasta que llegue un mensaje. Un ejemplo de uso, que deberia
     ser completado con el control de condiciones de error podria ser:

     \code

     void f (anna::ClientSocket* clientSocket)
        throw (anna::RuntimeException)
     {
        anna::Guard (clientSocket);

        if (clientSocket->wait (2000) == ClientSocket::Notify::ReceiveData) {
           const Message* data;

           while ((data = clientSocket->fetch ()) != NULL) {
              ...
              ..... procesa los datos ....
              ...
           }
        }
     }

     \endcode

     \param timeout Numero de milisegundos en que va a quedar bloqueado a la espera de
     obtener el primer mensaje.
     \param receive Un valor \em true indica que se tratara el mensaje detectado, un \em false
     indica que solo esperara la llegada de un bloque de datos, pero no se procesara en absoluto, solo
     devolvera Notify::ReceiveData.

     \return El resultado de la espera. En caso de no haber recibido ningun mensaje
     devolvera comm::Socket::Notify::None.

     \warning \li El hilo de ejecucion que invoque a este metodo queda bloqueado durante el
     tiempo que indica el 'timeout' o hasta que llegue un mensaje.
     \li La invocacion al metodo #receive y al #fetch deben estar protegidas por la misma
     seccion critica.
  */
  Notify::_v wait(const Millisecond &timeout, const bool receive = true) throw(RuntimeException);

  /**
     Elimina el contenido del buffer de recepcion.
  */
  void forgot() throw();

  /**
     Obtiene la capa de transporte asociada a este ClientSocket.
     \return la capa de transporte asociada a este ClientSocket.
  */
  Transport* getTransport() throw(RuntimeException) {
    return (a_transport != NULL) ? a_transport : reserveTransport();
  }

  /**
     Obtiene el receptor asociado a este ClientSocket. Puede ser NULL
     \return el receptor asociado a este ClientSocket.Puede ser NULL
     \warning Exclusivamente uso interno.
  */
  Receiver* getReceiver() throw(RuntimeException) {
    return (a_receiver != NULL) ? a_receiver : ((a_receiverFactory == NULL) ? NULL : reserveReceiver());
  }

  /**
   * Establece el receptor externo que tratara los mensajes recibidos por este socket.
   * Si el receptor se establece directamente por el usuario, sin tener una factoria de receptores
   * intermedia, invoca directamente al metodo anna::comm::Receiver::initialize.
   * \param receive La instancia del receptor externo que tratar los mensajes de este socket.
   */
  void setReceiver(Receiver* receive) throw(RuntimeException);

  /**
     Activa la solicitud de cierre del socket, que se llevara a cabo cuando el nucleo considere
     que se puede realizar de forma segura, desde el punto de vista de la informacion bloqueada
     por secciones criticas en los distintos componentes, que hacen uso del socket.
  */
  void requestClose() throw();

  /**
     Devuelve el estado de la solicitud de cierre del socket.
     \return el estado de la solicitud de cierre del socket.
  */
  bool hasRequestedClose() const throw() { return (a_status & Status::ClosePending) != 0; }

  /**
     Devuelve una cadena con la informacin referente a este socket.
     \return Una cadena con la informacin referente a este socket.
  */
  virtual std::string asString() const throw();

  /**
     Devuelve un nodo XML con la informacin referente a este objeto.
     \param parent Nodo XML a partir del cual introducir la informacin.
     \return Un nodo XML con la informacin referente a este objeto.
  */
  virtual xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

  /**
     Devuelve el nombre logico de esta clase.
     \return el nombre logico de esta clase.
  */
  static const char* className() throw() { return "anna::comm::ClientSocket"; }

protected:
  struct Status {
    enum _v { None = 0, Connected = 1, Corrupt = 2, ClosePending = 4, Working = 8 };

    static std::string asString(const int status) throw();
  };

  Transport* a_transport;
  AccessPoint a_remoteAccessPoint;

  /**
     Devuelve el numero de bytes contenido en el buffer intermedio de recepcion.
     \return El numero de bytes contenido en el buffer intermedio de recepcion.
  */
  int getBufferSize() const throw() { return a_data.getSize() - a_offset; }

  /**
     Devuelve el numero maximo de bytes que puede tener en el buffer intermedio de recepcion.
     \return el numero maximo de bytes que puede tener en el buffer intermedio de recepcion.
  */
  int getReceiveBufferSize() const throw() { return a_rcvBufferSize; }

  /**
     Obtiene los parametros de funcionamiento del Socket.
     \warning Exclusivamente uso interno.
  */
  void getSocketOptions() throw(RuntimeException);

  /**
     Recupera el ultimo mensaje recibido.
     El invocador debera haber establecido la forma de asegurar que hay que hay datos
      disponibles. Ademas debera establecer el tratamiento adecuado para cada uno
     de los posibles resultados.
     \return El resultado de la operacion.
     \warning La invocacion al metodo #receive y al #fetch deben estar protegidas por la
     misma seccion critica. Por ejemplo:

     \code

     void f (anna::ClientSocket* clientSocket)
        throw (anna::RuntimeException)
     {
        anna::Guard (clientSocket);

        if (clientSocket->receive () == ClientSocket::Notify::ReceiveData) {
           const Message* data;

           while ((data = clientSocket->fetch ()) != NULL) {
              ...
              ..... procesa los datos ....
              ...
           }
        }
     }

     \endcode
  */
  Notify::_v receive() throw(RuntimeException);

  /**
     Recupera el ultimo bloque de datos recuperado mediante el metodo #receive o #wait.
     El bloque de datos deberia ser interpretado segun las reglas de la capa de transporte
     asociada a este ClientSocket.

     \return El ultimo bloque recuperado mediante el metodo #receive o #wait. Puede ser NULL si no hay
     disponible ningun bloque.

     \warning Exclusivamente uso interno. Debe invocarse con una seccion
     critica activa sobre el ClientSocket.
  */
  const DataBlock* fetch() throw(RuntimeException);

  void activate(const Status::_v v) throw() { a_status |= v; }
  void deactivate(const Status::_v v) throw() { a_status &= ~v; }
  void deactivate(const int v) throw() { a_status &= ~v; }

  virtual int do_connect(const sockaddr*, const int len) throw(RuntimeException);
  virtual void do_write(const DataBlock&) throw(RuntimeException);
  virtual int do_read(const char* data, const int size) throw(RuntimeException);
  virtual void do_close() throw();

private:
  struct PendingBytes {
    Millisecond validUntil;
    int bytesToRead;

    PendingBytes() { validUntil = 0; bytesToRead = 0; }
  };

  //-----------------------------------------------------------------------------------------
  // a_status : Combinacion de bit's que indica el estado de la instancia.
  // a_expectedSize: Numero de bytes esperados. valdra -1 si todavia no hay informacion
  //                 para calcularlo.
  // a_data: Ultimo bloque de datos leido. Puede contener un numero indeterminado de mensajes
  // a_buffer: Referencia al espacio que ocupa el mensaje que estamos tratanto. Apunta
  // a un espacio dentro de 'a_data'-
  // a_reader: Buffer usado para leer los datos del Socket.
  // a_offset: Desplazamiento que hay que aplicar sobre los datos contenidos en 'a_data'.
  // a_cachedServer: Ultimo server calculado en el metodo getServer.
  //-----------------------------------------------------------------------------------------
  int a_status;
  int a_expectedSize;
  DataBlock a_data;
  Buffer a_buffer;
  DataBlock a_reader;
  int a_offset;
  comm::Server* a_cachedServer;
  Millisecond a_msMaxConnectionDelay;
  Millisecond a_msMaxWriteDelay;
  int a_rcvBufferSize;
  Receiver* a_receiver;
  mutable PendingBytes a_pendingBytes;
  bool a_ignoreIncomingMessages;

  void initialize() throw();
  void calculeExpectedSize(const DataBlock&) throw();
  Transport* reserveTransport() throw(RuntimeException);
  Transport* unsafe_reserveTransport() throw(RuntimeException);
  Receiver* reserveReceiver() throw(RuntimeException);
  int getTotalPendingBytes() const throw();

  friend class Communicator;
  // unsafe_reserveTransport


  friend class CongestionController;
  friend class handler::MetaClientSocket;
  friend class handler::DatagramSocket;
};

}
}

#endif

