// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_Device_hpp
#define anna_comm_Device_hpp

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string>

#include <anna/core/RuntimeException.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Network;

/**
   Abstraccion de un dispositivo de Red.
   Un Host puede tener asociado un numero indeterminado de dispositivos de red.

   El metodo Network::find(in_addr_t) creara un nuevo dispositivo de red.
*/
class Device {
public:
  /**
     Estados en los que puede estar un dispositivo de red.
  */
  struct Status { enum _v { Down, Up }; };

  /**
     Devuelve la direccion asociada a este instancia.
     \return La direccion asociada a este instancia.
  */
  in_addr_t getAddress() const throw() { return a_address; }

  /**
     Devuelve el estado asociado al dispositivo.
     \return el estado asociado al dispositivo.
  */
  Status::_v getStatus() const throw() { return a_status; }

  /**
     Establece el estado del dispositivo.
     \param status Nuevo estado del dispositivo.
  */
  void setStatus(const Status::_v status) throw() { a_status = status; }

  /**
    Operador de comparacion.
    \param right Direccion con la comparar.
    @return \em true si la direccion recibida como parametro coincide con esta.
    \em false en otro caso.
  */
  bool operator == (const Device& right) const throw()  { return a_address == right.a_address; }

  /**
    Operador de comparacion.
    \param right Direccion con la comparar.
    @return \em true si la direccion recibida como parametro coincide con esta.
    \em false en otro caso.
  */
  bool operator == (const in_addr_t& right) const throw() { return a_address == right; }

  /**
    Operador de comparacion.
    \param ip Direccion con la comparar. En formato A.B.C.D.
    @return \em true si la direccion recibida como parametro coincide con esta.
    \em false en otro caso.
  */
  bool operator == (const char* ip) const throw() { return a_address == inet_addr(ip); }

  /**
    Operador de comparacion.
    \param ip Direccion con la comparar. En formato A.B.C.D.
    @return \em true si la direccion recibida como parametro coincide con esta.
    \em false en otro caso.
  */
  bool operator == (const std::string& ip) const throw() { return a_address == inet_addr(ip.c_str()); }

  /**
    Operador distinto.
    @return \em true si la direccin recibida como par�etro coincide con �ta o
    \em false en otro caso.
  */
  bool operator != (const Device& right) const throw() { return a_address != right.a_address; }

  /**
    Operador distinto.
    @return \em true si la direccin recibida como par�etro no coincide con �ta o
    \em false en otro caso.
  */
  bool operator != (const in_addr_t right) const throw() { return a_address != right; }

  /**
     Devuelve una cadena la informacion mas relevante de esta instancia.
     @return una cadena la informacion mas relevante de esta instancia.
  */
  std::string asString() const throw();

  /**
     Devuelve un documento XML con la informacion mas relevante de esta instancia.
     \param parent Nodo XML del que deben depender los datos a crear.
     @return Un documento XML con la informacion mas relevante de esta instancia.
  */
  xml::Node* asXML(xml::Node* parent) const throw(RuntimeException);

  /**
     Incorpora los paremetros de esta instancia como atributos del nodo XML
     recibido como parametro.
     \param node Nodo del que dependen los atributos a crear.
  */
  void asAttribute(xml::Node* node) const throw(RuntimeException);

  /**
     Devuelve el nombre de esta clase. Se puede invocar desde \code template <class T>::asString (const T*); \endcode
     \return Devuelve el nombre de esta clase.
  */
  static const char* className() throw() { return "anna::comm::Device"; }

  /**
     Devuelve la direccion INET recibida como parametro en formato cadena.
     \param address Direccion INET a convertir.
     \return la direccion INET recibida como parametro en formato cadena.
  */
  static std::string asString(const in_addr_t& address) throw();

  /**
     Devuelve la direccin IP recibida como cadena en un tipo in_addr_t.
     \return la direccin IP recibida como cadena en un tipo in_addr_t.
  */
  static in_addr_t asAddress(const std::string& ip) throw() { return inet_addr(ip.c_str()); }

protected:
  /**
     Constructor.
     \param ip Texto con la direccin IP en formato A.B.C.D.
  */
  Device(const char* ip)  : a_address(inet_addr(ip)), a_status(Status::Up) {;}

  /**
     Constructor.
     \param ip Texto con la direccin IP en formato A.B.C.D.
  */
  Device(const std::string& ip)  : a_address(inet_addr(ip.c_str())), a_status(Status::Up) {;}

  /**
     Constructor.
     \param address Direccin IP de esta instancia
  */
  Device(const in_addr_t& address)  : a_address(address), a_status(Status::Up) {;}

  /**
     Constructor copia.
     \param other Dispositivo
  */
  Device(const Device& other) : a_address(other.a_address), a_status(Status::Up) {;}

private:
  const in_addr_t a_address;
  Status::_v a_status;

  friend class Network;
};

}
}

#endif



