// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_comm_ReceiverFactory_hpp
#define anna_comm_ReceiverFactory_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/mt/Mutex.hpp>
#include <anna/config/defines.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace comm {

class Receiver;

/**
   Interface for receivers factories.

   Is recommended (performance  issues) to use #Recycler<T> as in the following example:

   \code

   class MyReceiver;

   class MyReceiverFactory : public ReceiverFactory {
   public:
      MyReceiverFactory () : ReceiverFactory ("MyReceiverFactory") {;}

   private:
      anna::Recycler<MyReceiver> a_receivers;

      Receiver* do_create () throw () { return a_receivers.create (); }
      void do_release (Receiver* receiver) throw () { a_receivers.release (static_cast <MyReceiver*> (receiver)); }
   };

   \endcode

   In most of cases, is recommended to use an comm::ReceiverFactoryImpl instance with the receiver type used in our application.

   \see Receiver
   \see ReceiverFactoryImpl
*/
class ReceiverFactory : public Mutex {
public:
  /**
     Returns the logical name for this receiver factory.
     @return logical name for this receiver factory.
  */
  const std::string& getName() const throw() { return a_name; }

  /**
     Creates a receiver instance for this factory. Reuse resources when needed.

     \return The new receiver instance.
     \warning All the receivers obtained shall be released by mean #release.
  */
  Receiver* create() throw(RuntimeException);

  /**
     Release the receiver instance provided.

     \return The receiver instance to be released.
     \warning The parameter should have been obtained by mean #create.
  */
  void release(Receiver* receiver) throw();

  /**
     Returns a string with relevant information for this instance.
     @return string with relevant information for this instance.
  */
  std::string asString() const
  throw() {
    std::string msg("anna::comm::ReceiverFactory { Name: ");
    msg += a_name;
    return msg += " }";
  }

  /**
     Returns a XML document with relevant information for this instance.
     \param parent XML node from which created data will depend on.
     @return a XML document with relevant information for this instance.
  */
  xml::Node* asXML(xml::Node* parent) const throw();

protected:
  /**
     Constructor.
     \param name Logical name for this receivers factory.
  */
  ReceiverFactory(const char* name);

  /**
     Creates the receiver instance if no other reusable instance is found. This method is
     called from a critical section activated for this instance.

     \return The new receiver instance.
     \warning All the receivers obtained shall be released by mean #release.
  */
  virtual Receiver* do_create() throw() = 0;

  /**
     Release the receiver provided. This method is called from a critical section activated
     for this instance.

     \param The receiver instance to be released.
  */
  virtual void do_release(Receiver* receiver) throw() = 0;

private:
  const std::string a_name;

  ReceiverFactory(const ReceiverFactory&);

  WHEN_SINGLETHREAD(Receiver* a_receiver;)
};

}
}

#endif
