// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_DataBlock_hpp
#define anna_core_DataBlock_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/config/defines.hpp>

namespace anna {

/**
   Optimizacion de acceso a la memoria dinamica.

   Incrementa el rendimiento al acceder y reservar de memoria dinamica mediante la reutilizacion controlada.

   Para optimizar el acceso no se ha establecido ningun tipo de proteccion para ejecucion MT.
*/
class DataBlock {
public:
  /**
    Constructor.

    @param deepCopy Modo de copia de esta instancia. Si activamos el modo de copia profunda al asignar
    cualquier otro bloque de memoria a este, se reserva (si fuese necesario) la memoria para ubicar el
    buffer y despues realizara una copia byte a byte.
  */
  explicit DataBlock(const bool deepCopy = false) throw() :
    a_buffer(NULL),
    a_size(0),
    a_deepCopy(deepCopy),
    a_maxSize(0) {}

  /**
    Constructor.

    @param buffer Bloque de memoria con el que inicializar el buffer de esta instancia.
    @param size Numero de bytes del bloque de memoria recibido.
    @param deepCopy Modo de copia de esta instancia. Si activamos el modo de copia profunda al asignar
    cualquier otro bloque de memoria a este, se reserva la memoria para ubicar el buffer y despues
    realizara una copia byte a byte.
  */
  DataBlock(const char* buffer, const int size, const bool deepCopy = false) throw(RuntimeException);

  /**
    Constructor copia.
    El modo de copia sera el mismo que el establecido por la instancia de la que copiar.

    @param other Bloque de memoria con el que instanciar esta instancia.
  */
  DataBlock(const DataBlock& other) throw(RuntimeException);

  /**
     Destructor.
  */
  virtual ~DataBlock();

  // Accesores
  /**
     ste metodo solo debe usarse en aplicaciones mono-thread o en situaciones en las que estemos seguros
     que esta bloque de datos no puede verse afectado por otro thread.

     @return Tamao de la memoria reservada por esta instancia.
  */
  int getMaxSize() const throw() { return a_maxSize; }

  /**
     ste metodo solo debe usarse en aplicaciones mono-thread o en situaciones en las que estemos seguros
     que esta bloque de datos no puede verse afectado por otro thread.

     @return Tamao del bloque de memoria contenido actualmente.
  */
  int getSize() const throw() { return a_size; }

  /**
     ste metodo solo debe usarse en aplicaciones mono-thread o en situaciones en las que estemos seguros
     que esta bloque de datos no puede verse afectado por otro thread.

     @return El contenido del bloque de memoria.
  */
  const char* getData() const throw() { return a_buffer; }

  /**
     Devuelve informacion acerca del estado de ocupacion de este bloque de memoria.

    @return \em true si el bloque de memoria esta vacio o \em false en otro caso.
  */
  bool isEmpty() const throw() { return (a_size == 0) ? true : false; }

  /**
     Devuelve informacion acerca de la configuracion de reserva de memoria de este bloque.

     @return @em true si el bloque de memoria tiene activado el sistema de copia profunda o @em false en otro caso.
  */
  bool deepCopy() const throw() { return a_deepCopy; }

  /**
     Establece el numero de bytes que tiene asociado a este bloque de datos.
     \param size numero de bytes que tiene asociado a este bloque de datos.
     \warning El DataBlock delega la gestion de la memoria a la clase heredada.
  */
  void setSize(const int size) throw(RuntimeException);

  /**
    Anade el caracter recibido al bloque de memoria. Para poder usar este operador el bloque de
    memoria destino debe tener activado el modo de copia profunda. Si la memoria reservada no es
    suficiente  reservara automaticamente la cantidad de memoria necesaria.

    @param c Caracter a aadir a este bloque de memoria.

    @returns Una referencia a si mismo.
  */
  DataBlock& operator += (const char c) throw(RuntimeException) {
    append(&c, sizeof(c));
    return *this;
  }

  /**
    Anhade el bloque de memoria recibido. Para poder usar este operador el bloque de memoria
    destino debe tener activado el modo de copia profunda. Si la memoria reservada no es
    suficiente reservara automaticamente la cantidad de memoria necesaria.

    @param right Bloque de memoria a aadir a este bloque de memoria.

    @returns Una referencia a si mismo.
  */
  DataBlock& operator += (const DataBlock& right) throw(RuntimeException) {
    if(this != &right)
      append(right.a_buffer, right.a_size);

    return *this;
  }

  /**
    Anade la cadena recibida al bloque de memoria. Para poder usar este operador el bloque de
    memoria destino debe tener activado el modo de copia profunda. Si la memoria reservada no es
    suficiente  reservara automaticamente la cantidad de memoria necesaria.

    \param str Cadena a aadir a este bloque de memoria.

    @returns Una referencia a si mismo.
  */
  DataBlock& operator += (const std::string& str) throw(RuntimeException) {
    append(str.c_str(), str.length());
    return *this;
  }

  /**
     Devuelve la posicion i-esima del bloque de datos.
     \param pos Posicion a la que acceder.
     \return La posicion i-esima del bloque de datos.
  */
  const char operator [](const int pos) const throw(RuntimeException);

  /**
     Devuelve la posicion i-esima del bloque de datos.
     \param pos Posicion a la que acceder.
     \return La posicion i-esima del bloque de datos.
  */
  char& operator [](const int pos) throw(RuntimeException);

  /**
     Anade el bloque de memoria recibido. Para poder usar este operador el bloque de memoria
     destino debe tener activado el modo de copia profunda. Si la memoria reservada no es
     suficiente reservara automaticamente la cantidad de memoria necesaria.

     \param data Direccion donde comienza el bloque de datos.
     \param len Longitud del bloque de datos.
  */
  void append(const char* data, const int len) throw(RuntimeException);

  /**
     Anade el bloque de memoria recibido. Para poder usar este operador el bloque de memoria
     destino debe tener activado el modo de copia profunda. Si la memoria reservada no es
     suficiente reservara automaticamente la cantidad de memoria necesaria.

     \param other Bloque de memoria a aadir.
  */
  void append(const DataBlock& other) throw(RuntimeException) { append(other.a_buffer, other.a_size); }

  /**
    Copia el contenido del bloque recibido como parmetro.
    @param right Bloque de memoria del que copiar.
    @returns Una referencia a si mismo.
  */
  void assign(const DataBlock& right) throw(RuntimeException) { *this = right; }

  /**
    Copia o direcciona el contenido del bloque recibido como parmetro.
    \param buffer Direccin de memoria a direcionar.
    \param size Tamao de la memoria.
    @returns Una referencia a si mismo.
  */
  void assign(const char* buffer, const int size) throw(RuntimeException);

  /**
    operador copia. El modo de copiar vendra definido por el modo copia con el que hayamos
    iniciado la instancia destino.
    @param right Bloque de memoria del que copiar.
    @returns Una referencia a si mismo.
  */
  DataBlock& operator = (const DataBlock& right) throw(RuntimeException);

  /**
    Operador de inicializacion. El bloque destino debera tener activado el sistema de
    copia profunda.
    @param c Caracter con el que vamos a inicializar el contenido del bloque.
    @returns Una referencia a si mismo.
  */
  DataBlock& operator = (const char c) throw(RuntimeException) { clear(); (*this) += c; return *this; }

  /**
    Operador de inicializacion. El bloque destino debera tener activado el sistema de
    copia profunda.
    @param str Cadena con el que vamos a inicializar el contenido del bloque.
    @returns Una referencia a si mismo.
  */
  DataBlock& operator = (const std::string& str) throw(RuntimeException) { clear(); (*this) += str; return *this; }

  // Metodos
  /**
     Reserva el numero de bytes indicado por el parametro recibido. Si la cantidad de memoria preasignada es mayor
     que la solicitada no se realiza ninguna llamada al sistema operativo.

     @param nbytes Numero de bytes a reservar.
  */
  void allocate(const int nbytes) throw(RuntimeException);

  /**
    La reserva de memoria actual pasa a ser memoria pre-asignada, asi libera el bloque
    de memoria reservado hasta el momento, pero de forma que si posteriormente vuelve a
    ser necesario puede reutilizarlo sin tener que volver a realizar una llamada al
    sistema para obtener memoria dinamica.
  */
  void clear() throw(RuntimeException) { a_size = 0; }

  /**
     Elimina del bloque de memoria unos determinados bytes.

     @param pos Posicion del bloque donde empezar a eliminar.
     @param nbytes Numero de bytes a descartar a partir de la posicion indicada.
  */
  void remove(const int pos, const int nbytes) throw(RuntimeException);

  /**
     Elimina del bloque de memoria los n primeros bytes.
     @param nbytes Numero de bytes a descartar a partir de la posicion indicada.
  */
  void remove(const int nbytes) throw(RuntimeException);

  /**
   * Muestra el contenido de este buffer en forma de buffer hexadecimal vs bytes.
   */
  std::string asString(const int characterByLine = 24) const throw();

protected:
  /**
     Inicializa el contenido de este bloque de datos. Si fue instancia con copia
     profunda copia el contenido del buffer, en otro caso solo se queda con el
     valor de la referencia de memoria a la que apunta.

     @param buffer Bloque de memoria con el que inicializar el buffer de esta instancia.
     @param size Numero de bytes del bloque de memoria recibido.
  */
  void initialize(const char* buffer, const int size) throw(RuntimeException);

  /**
     Establece el espacio de memoria asociado a este bloque de datos.
     \param buffer Nuevo buffer de datos asociado a este bloque.
     \warning El DataBlock delega la gestion de la memoria a la clase heredada.
  */
  void setBuffer(const char* buffer) throw() { a_buffer = (char*) buffer; }

  /**
     Establece el numero de bytes que tiene reservados este bloque de datos.
     \param maxSize numero de bytes que tiene reservados este bloque de datos.
     \warning El DataBlock delega la gestion de la memoria a la clase heredada.
  */
  void setMaxSize(const int maxSize) throw() { a_maxSize = maxSize; }

private:
  char* a_buffer;
  int a_size;
  bool a_deepCopy;
  int a_maxSize;

  void extend(const int nbytes) throw(RuntimeException);
};

} //namespace anna

#endif

