// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_diameter_stack_Dictionary_hpp
#define anna_diameter_stack_Dictionary_hpp


// Local
#include <anna/diameter/defines.hpp>
#include <anna/diameter/stack/Format.hpp>
#include <anna/diameter/stack/Vendor.hpp>
#include <anna/diameter/stack/Avp.hpp>
#include <anna/diameter/stack/Command.hpp>

#include <anna/core/RuntimeException.hpp>

// STL
#include <string>
#include <map>

namespace anna {
namespace xml {
class Node;
class DTDMemory;
}
}




namespace anna {

namespace diameter {

namespace stack {



//------------------------------------------------------------------------------
//------------------------------------------------------------- class Dictionary
//------------------------------------------------------------------------------
/**
* Dictionary information
*/
class Dictionary {

public:

  typedef std::map<std::string, Format> format_container;
  typedef format_container::const_iterator const_format_iterator;

  typedef std::map<S32, Vendor> vendor_container;
  typedef vendor_container::const_iterator const_vendor_iterator;

  struct lessAvp { // order by vendor id
    bool operator()(AvpId id1, AvpId id2) const {
      if(id1.second == id2.second) return (id1.first < id2.first);

      return (id1.second < id2.second);
    }
  };
  typedef std::map<AvpId, Avp, lessAvp> avp_container;
  typedef avp_container::const_iterator const_avp_iterator;


  struct lessCommand {
    bool operator()(CommandId id1, CommandId id2) const {
      if(id1.first == id2.first) {
        if(id1.second != id2.second) return (id1.second);  // request -> answer

        return (false);
      }

      return (id1.first < id2.first);
    }
  };
  typedef std::map<CommandId, Command, lessCommand> command_container;
  typedef command_container::const_iterator const_command_iterator;

protected:
  std::string a_name;

private:

  bool a_allowUpdates;
  format_container a_formats;
  vendor_container a_vendors;
  avp_container a_avps;
  command_container a_commands;

  // Auxiliary
  const anna::xml::DTDMemory * a_dtd;

  // Name identifiers:
  typedef std::map<std::string, const Vendor *> vendorNames_container;
  typedef vendorNames_container::const_iterator const_vendorNames_iterator;

  typedef std::map<std::string, const Avp *> avpNames_container;
  typedef avpNames_container::const_iterator const_avpNames_iterator;

  typedef std::map<std::string, const Command *> commandNames_container;
  typedef commandNames_container::const_iterator const_commandNames_iterator;

  vendorNames_container a_vendorNames;
  avpNames_container a_avpNames;
  commandNames_container a_commandNames;


  // init
  void initialize() throw();

  // check & addings
  //void checkUniqueIdentifiers(const anna::xml::Node *rootNode) const throw(anna::RuntimeException);
  void extractFormats(const anna::xml::Node *rootNode) throw(anna::RuntimeException);
  void extractVendors(const anna::xml::Node *rootNode) throw(anna::RuntimeException);
  void extractAvps(const anna::xml::Node *rootNode) throw(anna::RuntimeException);
  void extractCommands(const anna::xml::Node *rootNode) throw(anna::RuntimeException);

public:

  Dictionary();
  ~Dictionary() {};

  // get
  const std::string & getName() const throw() { return a_name; }
  const Format * getFormat(const std::string & formatName) const throw();
  const Vendor * getVendor(S32 vendorId) const throw();
  const Vendor * getVendor(const std::string & vendorName) const throw();
  const Avp * getAvp(const AvpId & avpId) const throw();
  const Avp * getAvp(const std::string & avpName) const throw();
  const Command * getCommand(const CommandId & commandId) const throw();
  const Command * getCommand(const std::string & commandName) const throw();

  // set
  void allowUpdates(bool allow = true) throw() { a_allowUpdates = allow; }
  void addFormat(const Format &, bool reserved = false) throw(anna::RuntimeException);
  void addVendor(const Vendor &) throw(anna::RuntimeException);
  void addAvp(const Avp &) throw(anna::RuntimeException);
  void addCommand(const Command &) throw(anna::RuntimeException);

  // containers
  const_format_iterator format_begin() const throw() { return a_formats.begin(); }
  const_format_iterator format_end() const throw() { return a_formats.end(); }
  int format_size() const throw() { return a_formats.size(); }

  const_vendor_iterator vendor_begin() const throw() { return a_vendors.begin(); }
  const_vendor_iterator vendor_end() const throw() { return a_vendors.end(); }
  int vendor_size() const throw() { return a_vendors.size(); }

  const_avp_iterator avp_begin() const throw() { return a_avps.begin(); }
  const_avp_iterator avp_end() const throw() { return a_avps.end(); }
  int avp_size() const throw() { return a_avps.size(); }

  const_command_iterator command_begin() const throw() { return a_commands.begin(); }
  const_command_iterator command_end() const throw() { return a_commands.end(); }
  int command_size() const throw() { return a_commands.size(); }


  // helpers
  /**
  * Class string representation
  *
  * @return String with class content
  */
  std::string asString(void) const throw();

  /**
     Class XML representation.
     \param parent XML node over which we will put instance information.
     \return XML documentcon with class content.
  */
  anna::xml::Node* asXML(anna::xml::Node* parent) const throw();

  /**
     Class XML string representation
     \return XML string representation with class content.
  */
  std::string asXMLString() const throw();

  // operators

  /**
  * Loads an XML dictionary document over the diameter dictionary.
  *
  * Successive loadings will imply data accumulation, and the behaviour for redefinitions could be configured
  * in two ways: allow updating for dictionary items, or launch exception when any collision is found. This
  * could be set at #allowUpdates, and default value deny such redefinitions. It could be interesting, in order
  * to keep a more compact multi-stack configuration, allow updates sharing out the whole stack dictionaries as
  * many parts as possible to get centralized definitions. I.e. common vendors and avps used at a unique xml file,
  * and stack-specific commands in other set of xml files. Application only would have to load common part for all
  * the created dictionaries with this method and specific one for each final dictionary:
  *
  * @param pathFile Path file to the xml document which represents the diameter dictionary.
  *
  * <pre>
  * Dictionary *nokiaStack = stackEngine.createDictionary("/var/tmp/vendors_and_avps.xml", NOKIA_STACK_ID);
  * Dictionary *huaweiStack = stackEngine.createDictionary("/var/tmp/vendors_and_avps.xml", HUAWEI_STACK_ID);
  * nokiaStack->load("/var/tmp/nokia_commands.xml");
  * huaweiStack->load("/var/tmp/huawei_commands.xml");
  * </pre>
  */
  virtual void load(const std::string & pathFile) throw(anna::RuntimeException);

  /**
  * Clears dictionary content
  */
  void clear(void) throw() { initialize(); } // initialize and prepares the dictionary
};


}
}
}


#endif
