// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_ldap_Session_hpp
#define anna_ldap_Session_hpp

#include <string>

#include <anna/core/util/SortedVector.hpp>

#include <anna/comm/Handler.hpp>

#include <anna/ldap/defines.hpp>
#include <anna/ldap/ClassCode.hpp>

namespace anna {

namespace xml {
class Node;
}

namespace ldap {

class Request;
class Response;
class Timer;
class Engine;
class ResultCode;

/**
   Modela la conexion realizada contra una servidor LDAP.
*/
class Session : public comm::Handler {
public:
  /**
     Define los estados en los que puede estar una sesion LDAP.
     \see ldap::Session::getState
  */
  struct State {
    enum _v {
      Closed, /**< Cerrada */
      WaitingBind, /**< Pendiente de confirmacion de conexion */
      Bound /**< Conectada al servidor LDAP */
    };
  };

  /**
     Opciones disponibles para ajustar el comporamiento de nuestra sesion.
     \see ldap::Session
  */
  struct Option {
    /**
       Modos de interpretar las referencias.
       \see ldap::Session
    */
    struct Defer {
      enum _v {
        Never, /**< Valor por defecto */
        Searching, /**< Indica que los valores son interpretados durante la busqueda, pero no cuando localiza el objeto base de la misma */
        Finding, /**< Indica que los valores son interpretados cuando se localiza el objeto base, pero no durante la busqueda */
        Always
      };
    };
    /**
       Determina si la biblioteca LDAP realiza de forma automatica el siguimiento de las referencias retornadas
       por los servidores LDAP.
       \see ldap::Session
    */
    struct Referral { enum _v { Off, On }; };
  };

  /**
   * Periodo de espera por defecto para cada una de las peticiones de esta sesin.
   */
  static const Millisecond DefaultTimeout;

  /**
     Devuelve la direccion del servidor LDAP. Coincidira con la indicada en ldap::Engine::createSession.
     \return La direccion del servidor LDAP. Coincidira con la indicada en ldap::Engine::createSession.
  */
  const std::string& getURL() const throw() { return a_url; }

  /**
     Devuelve el usuario de esta sesion. Coincidira con la indicada en ldap::Engine::createSession.
     \return El usuario de esta sesion. Coincidira con la indicada en ldap::Engine::createSession.
  */
  const std::string& getUser() const throw() { return a_user; }

  /**
     Devuelve el password de esta sesion. Coincidira con la indicada en ldap::Engine::createSession.
     \return El password de esta sesion. Coincidira con la indicada en ldap::Engine::createSession.
  */
  const std::string& getPassword() const throw() { return a_password; }

  /**
     Devuelve la categoria de esta sesion. Coincidira con la indicada en ldap::Engine::createSession.
     \return La categoria de esta sesion. Coincidira con la indicada en ldap::Engine::createSession.
  */
  int getCategory() const throw() { return a_category; }

  /**
   * Devuelve la clave externa con la que fu creada esta sesin.
   */
  int getExternalID() const throw() { return a_externalID; }

  /**
     Devuelve el manejador usado para conectar con el servidor LDAP.
     \warning Exclusivamente uso interno.
  */
  void* getLDAP() throw() { return a_ldap; }

  /**
   * Devuelve el fd asociado a esta sesin de LDAP.
   * \return el fd asociado a esta sesin de LDAP.
   * \warning Las operaciones sobre este fd podran influir negativamente en el sistema.
   */
  int getDangerousFileDescriptor() const throw(RuntimeException);

  /**
     Devuelve el modo de interpretar las referencias establecido en esta sesion.
     \return El modo de interpretar las referencias establecido en esta sesion.
  */
  Option::Defer::_v getDefer() const throw() { return a_defer; }

  /**
     Devuelve \em true si la biblioteca LDAP realiza de forma automatica el siguimiento de las referencias
     retornadas por los servidores LDAP o \em false en otro caso.
     \return \em true si la biblioteca LDAP realiza de forma automatica el siguimiento de las referencias
     retornadas por los servidores LDAP o \em false en otro caso.
  */
  Option::Referral::_v getReferral() const throw() { return a_referral; }

  /**
     Devuelve el estado de esta sesion.
     \return El estado de esta sesion.
  */
  State::_v getState() const throw() { return a_state; }

  /**
   * Devuelve el periodo de espera establecido para las peticiones del tipo \em v.
   * \return el periodo de espera establecido para las peticiones del tipo \em v.
   */
  const Millisecond &getTimeout(const ClassCode::_v v) const throw() { return a_timeouts [v]; }

  /**
   * Devuelve el valor establecido en #setNetworkTimeout
   * \return Los miligundos establecidos por #setNetworkTimeout.
   * \warning El valor retornado slo tendr valided si #hasNetworkTimeout devolvi \em true.
   */
  Millisecond getNetworkTimeout() const throw() {
    Millisecond result(1000 * a_networkTimeout.tv_sec);
    return Millisecond(result + a_networkTimeout.tv_usec / 1000);
  }

  /**
   * Devuelve \em true si se estableci el tiempo de espera de conexin o \em false en otro caso.
   * \return \em true si se estableci el tiempo de espera de conexin o \em false en otro caso.
   */
  bool hasNetworkTimeout() const throw() { return a_networkTimeout.tv_sec != -1; }

  /**
     Solicita la conexion al servidor asociado a esta sesion.

     \warning Solo deberia invocarse directamente este metodo en caso de que el #eventResponse
     notifique no ha sido posible contactar con el servidor indicado.
  */
  void bind() throw(RuntimeException);

  /**
     Devuelve \em true si la sesion esta conectada al servidor LDAP o \em false en otro caso.
     \return \em true si la sesion esta conectada al servidor LDAP o \em false en otro caso.
  */
  bool isBound() const throw() { return a_state == State::Bound; }

  /**
     Establece el modo de interpretar las referencias.
     \param defer Indica el modo de interpretar las referencias.
  */
  void setOption(const Option::Defer::_v defer) throw() { a_defer = defer; }

  /**
     Establece el modo en que la biblioteca LDAP actua a la hora de realizar el siguimiento de las referencias.
     \param referral Indica el modo de realizar el seguimiento de las referncias.
  */
  void setOption(const Option::Referral::_v referral) throw() { a_referral = referral; }

  /**
   * Establece el periodo de tiempo mximo que esperar la conexin a un servidor LDAP antes de considerar que
   * ste no es alcanzable.
   *
   * \param timeout Milisegundos que esperar la conexin a un servidor LDAP antes de considerar que no es alcanzable.
   *
   * \see http://manpages.courier-mta.org/htmlman3/ldap_get_option.3.html
   * \warning Slo tiene efecto antes de invocar al #bind, por lo que habr que desactivar la auto conexin, para establecer el valor.
   */
  void setNetworkTimeout(const Millisecond &timeout) throw() { a_networkTimeout.tv_sec = timeout / 1000; a_networkTimeout.tv_usec = (timeout % 1000) * 1000; }

  /**
   * Elimina el tiempo asignado en #setNetworkTimeout.
   *
   * \see http://manpages.courier-mta.org/htmlman3/ldap_get_option.3.html
   * \warning Slo tiene efecto antes de invocar al #bind
   */
  void clearNetworkTimeout() throw() { a_networkTimeout.tv_sec = -1; a_networkTimeout.tv_usec = 0; }

  /**
     Establece el tiempo de espera maximo antes de considerar fallida una peticion LDAP.
     \param v Tipo de peticion LDAP.
     \param millisecond Milisegundos esperados antes de considerar fallida la peticion LDAP.

     Los temporizadores correspondientes las peticiones LDAP se activaran automaticamente al
     invocar a los distintos mtodos de esta clase.
  */
  void setTimeout(const ClassCode::_v v, const Millisecond &millisecond) throw() { a_timeouts [v] = millisecond;  }

  /**
     Envia la peticion recibida como parametro al servidor con el que estamos conectados mediante esta
     sesion LDAP.

     Una vez enviada la peticion se activara automaticamente un temporizador. Si este llegara a caducar
     se cancelara la busqueda y se invocara al metodo Session::eventResponse indicado que se ha producido
     un error de temporizacin. La duracion del temporizador sera la establecida por
     ldap::TimerManager::setTimeout o el valor defecto.

     \param request Peticion a enviar al servidor LDAP con el que estamos conectados.
     \return La instancia de la respuesta LDAP asociada la peticion realizada.
     \warning Solo se podra hacer uso de este metodo cuando el metodo #isBound devuelva \em true.
  */
  const Response* send(const Request* request) throw(RuntimeException);

  /**
     Envia la peticion recibida como parametro al servidor con el que estamos conectados mediante esta
     sesion LDAP.

     Una vez enviada la peticion se activara automaticamente un temporizador. Si este llegara a caducar
     se cancelara la busqueda y se invocara al metodo Session::eventResponse indicado que se ha producido
     un error de temporizacin. La duracion del temporizador sera la establecida por
     ldap::TimerManager::setTimeout o el valor defecto.

     \param request Peticion a enviar al servidor LDAP con el que estamos conectados.
     \return La instancia de la respuesta LDAP asociada la peticion realizada.
     \warning Solo se podra hacer uso de este metodo cuando el metodo #isBound devuelva \em true.
  */
  const Response* send(const Request& request) throw(RuntimeException) { return send(&request); }

  /**
   * Desconecta este cliente LDAP del servidor LDAP.
   * Se notifica la terminacin de cada una de las peticiones pendientes invocando al mtodo Session::eventResponse
   * indicando un error y se cancelan en el servidor LDAP.
   * \warning Despus de invocar a este mtodo habra que volver a invocar a ldap::Session::bind y esperar la conexin
   * antes de volver a usar esta sessin.
   */
  void unbind() throw(RuntimeException);

  /**
     Devuelve una cadena con la informacion relevante sobre esta instancia.
     \return Una cadena con la informacion relevante sobre esta instancia.
  */
  std::string asString() const throw();

  /**
     Devuelve un documento XML con la informacion relevante sobre esta instancia.
     \param parent Nodo XML del que colgar la informacion referente a esta instancia.
     \return Un documento XML con la informacion relevante sobre esta instancia.
  */
  xml::Node* asXML(xml::Node* parent) const throw();

protected:
  /**
     Constructor.
     \see ldap::Engine::createSession
  */
  Session();

  /**
     Metodo-manejador que informa de que el servidor LDAP con el que estaba conectado esta sesion
     ha dejado de dar servicio.

     Una vez que se ha notificado la caida de la sesion, el nucleo de ANNA.ldap genera un
     ldap::Session::eventResponse para cada una de las peticiones que hay pendientes de contestar.
  */
  virtual void eventServerShutdown() throw() {;}

  /**
     Metodo-manejador de las respuestas provenientes del servidor LDAP.

     \param response Objeto que contiene la respuesta correspondiente a la peticion LDAP realizada.
  */
  virtual void eventResponse(const Response& response) throw(RuntimeException) = 0;

  /**
     Metodo-manejador de los errores provenientes del servidor LDAP.
     \param resultCode Instancia que contiene la informacin sobre el error recibido.
     \param disconnect Incicador que informa al nivel de aplicacin sobre cmo actuar la ANNA.ldap
     para recuperar este error, si vale \em true la conexin se cerrar o \em false en otro caso.
  */
  virtual void eventResponseError(const ResultCode& resultCode, const bool disconnect) throw() {;}

  /**
   * Mtodo-manejador que se invoca cuando alguno de los mensajes intermedios requeridos para formar
   * la contestacin completa no se puede interpretar correctamente.
   *
     \param response Objeto que contiene la respuesta correspondiente a la peticion LDAP realizada.
   */
  virtual void eventIntermediateResponseError(const Response& response) throw() {;}

private:
  struct SortById {
    static IdMessage value(const Response*) throw();
  };
  typedef SortedVector <Response, SortById, IdMessage> response_container;
  typedef response_container::iterator response_iterator;
  typedef response_container::const_iterator const_response_iterator;

  typedef void* HandleLDAP;
  typedef void* HandleMessage;

  int a_category;
  HandleLDAP a_ldap;
  State::_v a_state;
  std::string a_url;
  std::string a_user;
  std::string a_password;
  Option::Defer::_v a_defer;
  Option::Referral::_v a_referral;
  response_container a_responses;
  int a_externalID;
  Millisecond a_timeouts [ClassCode::Max];
  struct timeval a_networkTimeout;

  /* Dependiendo de cmo se cree la sesin este miembro puede ser una copia de a_user
   * o el identificador numrico que pasaron como parmetro al crear la sesin
   */
  std::string a_keymap;

  void apply() throw(RuntimeException);
  void receiveBind(const IdMessage, HandleMessage) throw(RuntimeException);
  void receiveEntry(const IdMessage, HandleMessage) throw(RuntimeException);
  void receiveReference(const IdMessage, HandleMessage) throw(RuntimeException);
  void receiveResult(const IdMessage, HandleMessage) throw(RuntimeException);
  void expireResponse(Response*) throw();

  void finalize() throw();

  response_iterator response_begin() throw() { return a_responses.begin(); }
  response_iterator response_end() throw() { return a_responses.end(); }
  void response_add(Response* response) throw();
  void response_erase(Response* response) throw();
  Response* response_find(const IdMessage idMessage) throw(RuntimeException);

  static Response* response(response_iterator ii) throw() { return response_container::data(ii); }

  const_response_iterator response_begin() const throw() { return a_responses.begin(); }
  const_response_iterator response_end() const throw() { return a_responses.end(); }
  static const Response* response(const_response_iterator ii) throw() { return response_container::data(ii); }

  static const char* asText(const State::_v) throw();

  friend class Timer;
  friend class Engine;
};

}
}

#endif

