// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


// Local
#include <anna/diameter/stack/Engine.hpp>

#include <anna/xml/xml.hpp>
#include <anna/core/functions.hpp>
#include <anna/core/tracing/Logger.hpp>

#include <anna/core/functions.hpp>

// Standard
#include <stdarg.h>


namespace anna {
namespace diameter {
namespace stack {

const char *StackDTD = "\
<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\
<!-- Diameter dictionary DTD -->\n\
\n\
<!-- <!ELEMENT dictionary (format*, vendor+, avp+, command+)> allowed empty dictionary load: diameter base protocol would be harcoded -->\n\
<!ELEMENT dictionary (format*, vendor*, avp*, command*)>\n\
<!ATTLIST dictionary name CDATA #REQUIRED>\n\
<!--\n\
   Dictionary definitions\n\
\n\
   name: Dictionary name\n\
-->\n\
\n\
<!ELEMENT format EMPTY>\n\
<!ATTLIST format name CDATA #REQUIRED parent-type ( OctetString | Integer32 | Integer64 | Unsigned32 | Unsigned64 | Float32 | Float64 ) #REQUIRED>\n\
<!--\n\
   Format record\n\
\n\
   name:        Format type unique name\n\
   parent-type: Format parent type within diameter basics: OctetString, Integer32, Integer64, Unsigned32, Unsigned64, Float32, Float64\n\
-->\n\
\n\
<!ELEMENT vendor EMPTY>\n\
<!ATTLIST vendor name CDATA #REQUIRED code CDATA #REQUIRED>\n\
<!--\n\
   Vendor record\n\
\n\
   name: Vendor unique name\n\
   code: Vendor numeric code\n\
-->\n\
\n\
<!ELEMENT avp (single | grouped)>\n\
<!ATTLIST avp name CDATA #REQUIRED code CDATA #REQUIRED vendor-name CDATA #IMPLIED may-encrypt (yes | no) #IMPLIED v-bit (must | may | shouldnot | mustnot) #IMPLIED m-bit (must | may | shouldnot | mustnot) #IMPLIED p-bit (must | may | shouldnot | mustnot) #IMPLIED>\n\
<!--\n\
   Avp record\n\
\n\
   name:          Avp unique name\n\
   code:          Avp numeric code\n\
   vendor-name:   Avp vendor name with vendor-id value. Gets IETF (zero value) when missing\n\
   may-encrypt:   Avp data may be encrypted (yes, no). By default is 'no'\n\
   v-bit:         Avp vendor-specific bit rule (must, may, shouldnot, mustnot). By default is 'mustnot'\n\
   m-bit:         Avp mandatory bit rule (must, may, shouldnot, mustnot). By default is 'may'\n\
   p-bit:         Avp protected (encryption) bit rule (must, may, shouldnot, mustnot). By default is 'may'\n\
-->\n\
\n\
<!ELEMENT single (label*)>\n\
<!ATTLIST single format-name CDATA #REQUIRED enum CDATA #IMPLIED>\n\
<!--\n\
   Non-grouped avp struct\n\
\n\
   format-name: Avp format type name (RFC3588: 'OctetString, Integer32, Integer64, Unsigned32, Unsigned64, Float32, Float64, Address, Time, UTF8String, DiameterIdentity, DiameterURI, Enumerated, IPFilterRule, QoSFilterRule' or application format data type names)\n\
   enum:        Enum values restriction for enumerated-format avp: '[i][,[j]][,[m-n]...]'\n\
-->\n\
\n\
<!ELEMENT label EMPTY>\n\
<!ATTLIST label data CDATA #REQUIRED alias CDATA #REQUIRED>\n\
<!--\n\
   Mostly used with enumerated-type avps\n\
\n\
   data:    Specific avp data\n\
   alias:   Data-linked name\n\
-->\n\
\n\
<!ELEMENT grouped (avprule+)>\n\
<!--\n\
   Grouped avp struct\n\
-->\n\
\n\
<!ELEMENT avprule EMPTY>\n\
<!ATTLIST avprule id CDATA #REQUIRED type (Fixed | Mandatory | Optional) #REQUIRED qual CDATA #IMPLIED>\n\
<!--\n\
   Avp rule definition\n\
\n\
   id:   Avp name reference. Generic AVP is harcoded with 'AVP' name\n\
   type: Avp type (Fixed, Mandatory, Optional)\n\
   qual: Avp qualifier: *:0-inf; x*:x-inf; *x:0-x; x*y:x-y; missing:1(Fixed, Mandatory), 0-1(Optional)\n\
-->\n\
\n\
<!ELEMENT command (avprule*)>\n\
<!ATTLIST command name CDATA #REQUIRED code CDATA #REQUIRED type (Request | Answer) #REQUIRED>\n\
<!--\n\
   Command record\n\
\n\
   name:       Command unique name\n\
   code:       Command numeric code\n\
   type:       Command type (Request, Answer)\n\
-->\n\
\n\
";

}
}
}


//------------------------------------------------------------------------------
//------------------------------------------------------------- Engine::Engine()
//------------------------------------------------------------------------------
anna::diameter::stack::Engine::Engine(void) {
  anna::xml::functions::initialize();
  a_dtd.initialize(StackDTD);
}


//------------------------------------------------------------------------------
//------------------------------------------------------ Engine::getDictionary()
//------------------------------------------------------------------------------
const anna::diameter::stack::Dictionary * anna::diameter::stack::Engine::getDictionary(unsigned int stackId) const throw() {
  const Dictionary * result = NULL;
  const_stack_iterator it = a_stacks.find(stackId);

  if(it != stack_end()) result = (*it).second;

  return result;
}


//------------------------------------------------------------------------------
//----------------------------------------------------------- Engine::asString()
//------------------------------------------------------------------------------
std::string anna::diameter::stack::Engine::asString(bool all) const throw() {
  std::string trace;

  if(isEmpty()) {
    trace = "No diameter stacks found";
  } else {
    int numberOfStacks = stack_size();
    trace = ((numberOfStacks > 1) ? "Multi-stack " : "Mono-stack ");
    trace += "configuration, ";
    trace += anna::functions::entriesAsString(numberOfStacks);
    trace += ":\n";
    trace += "\n";

    for(const_stack_iterator it = stack_begin(); it != stack_end(); it++) {
      std::string title = "Diameter stack id = ";
      title += anna::functions::asString((*it).first);
      trace += anna::functions::highlightJustify(title);

      if(all) trace += (*it).second->asString();
      else trace += (*it).second->getName();

      trace += "\n";
    }
  }

  trace += "\n";
  return (trace);
}


//------------------------------------------------------------------------------
//------------------------------------------------- Engine::registerDictionary()
//------------------------------------------------------------------------------
anna::diameter::stack::Dictionary * anna::diameter::stack::Engine::registerDictionary(unsigned int stackId, Dictionary *dictionary) throw(anna::RuntimeException) {
  Dictionary * result = const_cast<Dictionary *>(getDictionary(stackId));

  if(!dictionary)
    throw anna::RuntimeException("Cannot provide a NULL dictionary. It must be previously allocated", ANNA_FILE_LOCATION);

  if(result) {  // if exists, launch exception
    throw anna::RuntimeException("Such provided stack id has already been created. Removes it before call this method", ANNA_FILE_LOCATION);
  } else { // new stack
    a_stacks[stackId] = dictionary; // no need for singleton destructor
    //const_stack_iterator it = a_stacks.find(stackId);
    //result = (Dictionary *)(*it).second;
    result = dictionary;
  }

  return result;
}

//------------------------------------------------------------------------------
//--------------------------------------------------- Engine::createDictionary()
//------------------------------------------------------------------------------
anna::diameter::stack::Dictionary *  anna::diameter::stack::Engine::createDictionary(unsigned int stackId, const std::string & xmlPathFile) throw(anna::RuntimeException) {
  Dictionary * result = const_cast<Dictionary *>(getDictionary(stackId));

  if(result)  // if exists, launch exception
    throw anna::RuntimeException("Such provided stack id has already been created (note: API allows you to remove any registered stack)", ANNA_FILE_LOCATION);

  // Register a new dictionary:
  result = registerDictionary(stackId, new Dictionary());

  if(xmlPathFile != "") {
    try {
      result->load(xmlPathFile);
    } catch(anna::RuntimeException& ex) {
      ex.trace();
      throw ex;
    }
  }

  return result;
}


void anna::diameter::stack::Engine::loadDictionary(const std::vector<unsigned int> & stacks, const std::string & xmlPathFile) throw(anna::RuntimeException) {
  std::vector<unsigned int>::const_iterator it;
  Dictionary *d;

  if(xmlPathFile == "")
    throw anna::RuntimeException("Empty xml path file provided", ANNA_FILE_LOCATION);

  for(it = stacks.begin(); it != stacks.end(); it++) {
    d = const_cast<Dictionary*>(getDictionary(*it));

    if(d) d->load(xmlPathFile);
    else {
      LOGWARNING(
        std::string trace = "Cannot load dictionary '";
        trace += xmlPathFile;
        trace += "' over stack id '";
        trace += anna::functions::asString(*it);
        trace += "' because it doesn't exists";
        anna::Logger::warning(trace, ANNA_FILE_LOCATION);
      );
    }
  }
}

void anna::diameter::stack::Engine::loadDictionary(const std::string & xmlPathFile) throw(anna::RuntimeException) {
  Dictionary *d;

  if(xmlPathFile == "")
    throw anna::RuntimeException("Empty xml path file provided", ANNA_FILE_LOCATION);

  for(const_stack_iterator it = stack_begin(); it != stack_end(); it++) {
    d = const_cast<Dictionary*>(getDictionary((*it).first));
    d->load(xmlPathFile);
  }
}


//------------------------------------------------------------------------------
//-------------------------------------------------------- Engine::removeStack()
//------------------------------------------------------------------------------
void anna::diameter::stack::Engine::removeStack(unsigned int stackId) throw() {
  stack_iterator it = a_stacks.find(stackId);

  if(it != stack_end()) {  // if exists, clear
    //(*it).second->clear();
    a_stacks.erase(it);
  } else { // new stack
    LOGWARNING(
      std::string trace = "Cannot remove stack id '";
      trace += anna::functions::asString(stackId);
      trace += "' because it doesn't exists";
      anna::Logger::warning(trace, ANNA_FILE_LOCATION);
    );
  }
}
